#import "XeeDirectoryController.h"
#import "XeeView.h"
#import "XeeImage.h"
#import "XeeDelegate.h"
#import "XeeDestinationList.h"
#import "XeeMisc.h"
#import "XeeStrings.h"



// XeeDirectoryController

@implementation XeeDirectoryController

static NSMutableArray *directorycontrollers=nil;

-(id)init
{
	if(self=[super init])
	{
		directory=nil;
		dircontents=nil;

		loadlock=[[NSRecursiveLock alloc] init];
		loader_running=NO;
		exiting=NO;

		previmage=nil;
		currimage=nil;
		nextimage=nil;
		loadingimage=nil;

		previndex=currindex=nextindex=-1;

		randomlist=NULL;

		drawer_mode=pending_mode=XeeNoMode;

		[directorycontrollers addObject:self];
    }
    return self;
}

-(void)dealloc
{
	//[controllers removeObject:self];

	[directory release];
	[dircontents release];

	[loadlock release];

	[previmage release];
	//[currimage release];
	[nextimage release];

	free(randomlist);

	[drawer release];
	[renamepanel release];
	[collisionpanel release];

	[super dealloc];
}

-(void)windowWillClose:(NSNotification *)notification
{
	[directorycontrollers removeObject:self];

	exiting=YES;
	[loadingimage stopLoading];

	[super windowWillClose:notification];
}

-(NSRect)availableScreenSpace
{
	NSRect rect=[super availableScreenSpace];
	if([drawer state]==NSDrawerOpenState) rect.size.width-=[drawer contentSize].width+6;
	return rect;
}

-(NSSize)minViewSize
{
	NSSize size=[super minViewSize];

	if([drawer state]==NSDrawerOpenState)
	{
		int drawerheight=[table numberOfRows]*18;
		drawerheight+=[drawer leadingOffset];
		drawerheight+=[drawer trailingOffset];
		drawerheight+=19; // uh-huh, right...
		if(size.height<drawerheight) size.height=drawerheight;
	}
	return size;
}



-(BOOL)loadImage:(NSString *)filename
{
	BOOL dir;

	if(![[NSFileManager defaultManager] fileExistsAtPath:filename isDirectory:&dir]) return NO;

	[loadlock lock];

	[self setPreviousImage:nil index:-1];
	[self setCurrentImage:nil index:-1];
	[self setNextImage:nil index:-1];
	[loadingimage stopLoading];

	[loadlock unlock];

	free(randomlist);
	randomlist=NULL;

	NSString *filepart;

	if(dir)
	{
		[directory release];
		directory=[filename retain];
		filepart=nil;
	}
	else
	{
		[directory release];
		directory=[[filename stringByDeletingLastPathComponent] retain];
		filepart=[filename lastPathComponent];
	}

	[dircontents release];
	dircontents=[[NSMutableArray alloc] initWithCapacity:256];

	NSFileManager *fm=[NSFileManager defaultManager];
	NSArray *filetypes=[XeeImage fileTypes];
	NSArray *files=[fm directoryContentsAtPath:directory];

	NSEnumerator *enumerator=[files objectEnumerator];
	NSString *file;

	while(file=[enumerator nextObject])
	{
		NSDictionary *attrs=[fm fileAttributesAtPath:[directory stringByAppendingPathComponent:file] traverseLink:YES];
		NSString *type=NSFileTypeForHFSTypeCode([attrs fileHFSTypeCode]);
		NSString *ext=[[file pathExtension] lowercaseString];

		if( (filepart&&[file isEqual:filepart])
		||[filetypes indexOfObject:ext]!=NSNotFound
		||[filetypes indexOfObject:type]!=NSNotFound)
		[dircontents addObject:file];
	}

	[self sortFiles];

	int index;

	if(filepart)
	{
		index=[self findFile:filepart];
	}
	else
	{
		if([dircontents count]) index=0;
		else index=-1;
	}

	[self displayImage:index next:index+1];

	return index!=-1;
}

-(void)displayImage:(int)index next:(int)next
{
	[loadlock lock];

	if(index<0 || index>=[dircontents count]) index=-1;
	if(next<0 || next>=[dircontents count]) next=-1;

	XeeImage *newcurrimage=[self imageAtIndex:index];
	XeeImage *newnextimage=[self imageAtIndex:next];

	[self setPreviousImage:currimage index:currindex];
	[self setCurrentImage:newcurrimage index:index];
	[self setNextImage:newnextimage index:next];

	if(loadingimage&&loadingimage!=currimage) [loadingimage stopLoading];

	[self launchLoader];

	[loadlock unlock];

	[window setTitleWithRepresentedFilename:[self currentFilename]];
}

-(XeeImage *)imageAtIndex:(int)index
{
	if(index<0) return nil;
	else if(index==currindex) return currimage;
	else if(index==nextindex) return nextimage;
	else if(index==previndex) return previmage;
	else return [XeeImage imageForFilename:[directory stringByAppendingPathComponent:[dircontents objectAtIndex:index]]];
}

-(void)setCurrentImage:(XeeImage *)image index:(int)index
{
	currindex=index;
	[self setImage:image];
}

-(void)setPreviousImage:(XeeImage *)image index:(int)index
{
	previndex=index;
	[previmage autorelease];
	previmage=[image retain];
}

-(void)setNextImage:(XeeImage *)image index:(int)index
{
	nextindex=index;
	[nextimage autorelease];
	nextimage=[image retain];
}

-(void)reloadImage
{
	[loadlock lock];

	[self setImage:[XeeImage imageForFilename:[directory stringByAppendingPathComponent:[dircontents objectAtIndex:currindex]]]];

	[loadingimage stopLoading];
	[self launchLoader];

	[loadlock unlock];
}

-(void)imageLoader:(id)nothing
{
	NSAutoreleasePool *pool=[[NSAutoreleasePool alloc] init];

	[self retain];

	[NSThread setThreadPriority:0.1];

	[loadlock lock];

	for(;;)
	{
		if(exiting) break;

		NSAutoreleasePool *pool=[[NSAutoreleasePool alloc] init];

		if(currimage&&![currimage loaded]) loadingimage=currimage;
		else if(nextimage&&![nextimage loaded]) loadingimage=nextimage;
		else break;

		[loadingimage retain];
		[loadlock unlock];

		[loadingimage load];

		[loadlock lock];
		[loadingimage release];

		[pool release];
	}

	loader_running=NO;
	loadingimage=nil;

	[loadlock unlock];

	[self release];

	[pool release];
}

-(void)launchLoader
{
	if(!loader_running)
	{
		[NSThread detachNewThreadSelector:@selector(imageLoader:) toTarget:self withObject:nil];
		loader_running=YES;
	}
}

-(void)updateStatusBar
{
	if(currindex>=0)
	{
		NSString *filepart=[dircontents objectAtIndex:currindex];

		if(currimage)
		{
			NSString *status=[NSString stringWithFormat:@"%d/%d: %@ (%@ %dx%d %@, %@, %d of %d frames)",
			currindex+1,[dircontents count],filepart,[currimage descriptiveFilesize],
			[currimage width],[currimage height],[currimage format],[currimage descriptiveDepth],[currimage frame]+1,
			[currimage frames]];

			[statusbar setStringValue:status];
		}
		else
		{
			NSString *status=[NSString stringWithFormat:@"%d/%d - %@ - error loading image",currindex+1,[dircontents count],filepart];
			[statusbar setStringValue:status];
		}
	}
	else
	{
		[statusbar setStringValue:@"No file loaded"];
	}
}



int file_compare(NSString *a,NSString *b,void *context) { return [a compare:b options:NSCaseInsensitiveSearch|NSNumericSearch]; }

-(void)sortFiles
{
	[dircontents sortUsingFunction:file_compare context:nil];
}

-(int)findFile:(NSString *)filename
{
	if(!filename) return -1;

	int index=[dircontents indexOfObject:[filename lastPathComponent]];
	if(index==NSNotFound) return -1;
	return index;
}

-(int)findNextFile:(NSString *)filename
{
	if(!filename) return 0;

	int count=[dircontents count];

	for(int i=0;i<count;i++)
	{
		if(file_compare(filename,[dircontents objectAtIndex:i],nil)==NSOrderedAscending) return i;
	}

	return count-1;
}

-(void)removeFile:(NSString *)filename
{
	int index=[self findFile:filename];
	if(index<0) return;

	[loadlock lock];

	NSString *prevname=previndex>=0?[[dircontents objectAtIndex:previndex] retain]:nil;
	NSString *currname=currindex>=0?[[dircontents objectAtIndex:currindex] retain]:nil;
	NSString *nextname=nextindex>=0?[[dircontents objectAtIndex:nextindex] retain]:nil;

	[dircontents removeObjectAtIndex:index];

	previndex=[self findFile:prevname];
	currindex=[self findFile:currname];
	nextindex=[self findFile:nextname];

	if(previndex<0) { [previmage release]; previmage=nil; }
	if(nextindex<0) { [nextimage release]; nextimage=nil; }
	if(currindex<0)
	{
		[self displayImage:[self findNextFile:currname] next:nextindex];
	}

	[prevname release];
	[currname release];
	[nextname release];

	[loadlock unlock];
}

-(void)insertFile:(NSString *)filename
{
	if(![[filename stringByDeletingLastPathComponent] isEqual:directory]) return; // wrong directory
	if([self findFile:filename]>=0) return; // already there

	[loadlock lock];

	NSString *prevname=previndex>=0?[[dircontents objectAtIndex:previndex] retain]:nil;
	NSString *currname=currindex>=0?[[dircontents objectAtIndex:currindex] retain]:nil;
	NSString *nextname=nextindex>=0?[[dircontents objectAtIndex:nextindex] retain]:nil;

	[dircontents addObject:[filename lastPathComponent]];
	[self sortFiles];

	previndex=[self findFile:prevname]; [prevname release];
	currindex=[self findFile:currname]; [currname release];
	nextindex=[self findFile:nextname]; [nextname release];

	[loadlock unlock];

	[self displayImage:currindex next:nextindex];
}



-(NSDrawer *)drawer { return drawer; }

-(NSString *)directory { return directory; }

-(NSString *)currentFilename
{
	if(currindex<0) return nil;
	return [directory stringByAppendingPathComponent:[dircontents objectAtIndex:currindex]];
}



-(IBAction)skipNext:(id)obj
{
	[self skip:1];
}

-(IBAction)skipPrev:(id)obj
{
	[self skip:-1];
}

-(IBAction)skipFirst:(id)obj
{
	[self displayImage:0 next:1];
}

-(IBAction)skipLast:(id)obj
{
	[self displayImage:[dircontents count]-1 next:[dircontents count]-2];
}

-(void)skipToFile:(NSString *)filename
{
	int index=[self findFile:filename];

	if(index>=0) [self displayImage:index next:index+1];
}

-(void)skip:(int)step
{
	if(currindex==-1)
	{
		if(step>=0) [self skipFirst:nil];
		else [self skipLast:nil];
	}
	else
	{
		int newindex=currindex+step;
		int newnext=currindex+step+(step>=0?1:-1);
		int count=[dircontents count];

		if(newindex<0) newindex=0;
		if(newindex>=count) newindex=count-1;
		if(newnext<0) newnext=-1;
		if(newnext>=count) newnext=-1;

		if(newindex==currindex) return;

		[self displayImage:newindex next:newnext];
	}
}

-(IBAction)skipRandom:(id)sender
{
	if(currindex<0) return;
	if(!randomlist) [self buildRandomList];

	int newindex=randomlist[2*currindex];
	int newnext=randomlist[2*newindex];

	[self displayImage:newindex next:newnext];
}

-(IBAction)skipRandomPrev:(id)sender
{
	if(currindex<0) return;
	if(!randomlist) [self buildRandomList];

	int newindex=randomlist[2*currindex+1];
	int newnext=randomlist[2*newindex+1];

	[self displayImage:newindex next:newnext];
}

-(void)buildRandomList
{
	free(randomlist);

	srandom(time(NULL));

	int length=[dircontents count];

	int *order=malloc(sizeof(int)*length);
	randomlist=malloc(2*sizeof(int)*length);

	for(int i=0;i<length;i++) order[i]=i;

	for(int i=length-1;i>0;i--)
	{
		int randindex=random()%i;
		int tmp=order[i];
		order[i]=order[randindex];
		order[randindex]=tmp;
	}

	for(int i=0;i<length;i++)
	{
		randomlist[2*order[i]]=order[(i+1)%length];
		randomlist[2*order[i]+1]=order[(i+length-1)%length];
	}

	free(order);
}



-(IBAction)revealInFinder:(id)sender
{
	if(currindex<0) return;

	NSString *filename=[self currentFilename];

	[[NSWorkspace sharedWorkspace] selectFile:filename inFileViewerRootedAtPath:nil];
}



-(IBAction)renameFileFromMenu:(id)sender
{
	if(currindex<0) return;

	[renamepanel run:window image:currimage];
}

-(void)renameFile:(NSString *)filename to:(NSString *)newname
{
	if([filename isEqual:newname]) return;

	if([[NSFileManager defaultManager] fileExistsAtPath:newname])
	{
		[self errorMessage:S_RENAMEERRORTITLE text:[NSString stringWithFormat:S_RENAMECOLLISIONCONTENT,[filename lastPathComponent]]];
	}
	else
	{
		if([[NSFileManager defaultManager] movePath:filename toPath:newname handler:nil])
		{
			[self insertFile:newname];
			[self skipToFile:newname];
			[self removeFile:filename];
		}
		else
		{
			[self errorMessage:S_RENAMEERRORTITLE text:[NSString stringWithFormat:S_RENAMEERRORCONTENT,[filename lastPathComponent]]];
		}
	}
}


-(IBAction)deleteFileFromMenu:(id)sender
{
	if(currindex<0) return;

	[self deleteFile:[self currentFilename]];
}

-(IBAction)askAndDelete:(id)sender
{
	if(currindex<0) return;

	NSString *filename=[[self currentFilename] retain];
	NSAlert *alert=[[NSAlert alloc] init];

	[alert setMessageText:S_DELETETITLE];
	[alert setInformativeText:[NSString stringWithFormat:S_DELETECONTENT,[filename lastPathComponent]]];
	[alert addButtonWithTitle:S_DELETE];
	[alert addButtonWithTitle:S_CANCEL];
	[alert setIcon:[[[NSImage alloc] initWithContentsOfFile:@"/System/Library/CoreServices/Dock.app/Contents/Resources/trashfull.png"] autorelease]];

	[alert beginSheetModalForWindow:window modalDelegate:self didEndSelector:@selector(deleteAlertEnd:returnCode:contextInfo:) contextInfo:[filename retain]];
}

-(void)deleteAlertEnd:(NSAlert *)alert returnCode:(int)res contextInfo:(NSString *)filename
{
	if(res==NSAlertFirstButtonReturn)
	{
		[self deleteFile:filename];
	}
	[filename release];
}

-(void)deleteFile:(NSString *)filename
{
	if([[NSWorkspace sharedWorkspace] performFileOperation:NSWorkspaceRecycleOperation source:[filename stringByDeletingLastPathComponent]
	destination:nil files:[NSArray arrayWithObject:[filename lastPathComponent]] tag:nil])
	{
		[self removeFile:filename];
		[[[[NSSound alloc] initWithContentsOfFile:@"/System/Library/Components/CoreAudio.component/Contents/Resources/SystemSounds/dock/drag to trash.aif" byReference:NO] autorelease] play];
	}
	else
	{
		[self errorMessage:S_DELERRORTITLE text:[NSString stringWithFormat:S_DELERRORCONTENT,[filename lastPathComponent]]];
	}
}



-(IBAction)moveFile:(id)sender
{
	[self triggerDrawer:XeeMoveMode];
}

-(IBAction)copyFile:(id)sender
{
	[self triggerDrawer:XeeCopyMode];
}

-(void)triggerDrawer:(int)mode
{
	int state=[drawer state];
	if(state==NSDrawerClosedState)
	{
		if(mode==XeeMoveMode) [drawertext setStringValue:S_MOVETITLE];
		else if(mode==XeeCopyMode) [drawertext setStringValue:S_COPYTITLE];

		[drawer openOnEdge:NSMaxXEdge];
		drawer_mode=mode;
	}
	else
	{
		if(state==NSDrawerOpenState||state==NSDrawerOpeningState) [drawer close];
		else if(mode==drawer_mode) [drawer open];

		if(drawer_mode!=mode) pending_mode=mode;
	}
}

-(void)drawerWillOpen:(NSNotification *)notification
{
	[window makeFirstResponder:table];
	[[drawer contentView] setNextResponder:window];
}

-(void)drawerDidClose:(NSNotification *)notification
{
	if(pending_mode!=XeeNoMode)
	{
		[self triggerDrawer:pending_mode];
		pending_mode=XeeNoMode;
	}
}

-(void)destinationListClick:(id)sender
{
	if(currindex<0) return;

	NSString *filename=[self currentFilename];
	int index=[sender selectedRow];

	if(index==0)
	{
		NSOpenPanel *panel=[NSOpenPanel openPanel];
		[panel setCanChooseDirectories:YES];
		[panel setCanChooseFiles:NO];
		[panel setCanCreateDirectories:YES];
		if(drawer_mode==XeeMoveMode) [panel setPrompt:S_MOVE];
		else if(drawer_mode==XeeCopyMode) [panel setPrompt:S_COPY];

		[panel beginSheetForDirectory:nil file:nil types:nil modalForWindow:window modalDelegate:self
		didEndSelector:@selector(destinationPanelEnd:returnCode:contextInfo:) contextInfo:[filename retain]];
	}
	else
	{
		NSString *destination=[[[[sender dataSource] destination:index-1] objectForKey:@"path"] stringByAppendingPathComponent:[filename lastPathComponent]];
		[self attemptToTransferFile:filename to:destination mode:drawer_mode];
	}
}

-(void)destinationPanelEnd:(NSOpenPanel *)panel returnCode:(int)res contextInfo:(NSString *)filename
{
	if(res==NSOKButton)
	{
		NSString *destdir=[[panel filenames] objectAtIndex:0];
		NSString *destination=[destdir stringByAppendingPathComponent:[filename lastPathComponent]];

		[[table dataSource] suggestInsertion:destdir];

		[self attemptToTransferFile:filename to:destination mode:drawer_mode];
	}
	[filename release];
}

-(void)attemptToTransferFile:(NSString *)filename to:(NSString *)destination mode:(int)mode
{
	if([filename isEqual:destination])
	{
		[self errorMessage:S_TRANSFERSAMETITLE
		text:[NSString stringWithFormat:S_TRANSFERSAMECONTENT,[filename lastPathComponent]]];
	}
	else
	{
		NSDictionary *destinfo=[[NSFileManager defaultManager] fileAttributesAtPath:destination traverseLink:YES];

		if(destinfo)
		{
			XeeImage *destimage=[XeeImage imageForFilename:destination];
			[collisionpanel run:window source:currimage destination:destimage mode:drawer_mode];
		}
		else
		{
			[self transferFile:filename to:destination mode:drawer_mode];
		}
	}
}

-(void)transferFile:(NSString *)filename to:(NSString *)destination mode:(int)mode
{
	if([[NSFileManager defaultManager] fileExistsAtPath:destination])
	{
		[[NSFileManager defaultManager] removeFileAtPath:destination handler:nil];
	}

	if(mode==XeeMoveMode)
	{
		if([[NSFileManager defaultManager] movePath:filename toPath:destination handler:nil])
		{
			// "moved" message in status bar
			[[[[NSSound alloc] initWithContentsOfFile:@"/System/Library/Components/CoreAudio.component/Contents/Resources/SystemSounds/system/Volume Mount.aif" byReference:NO] autorelease] play];
			[self removeFile:filename];
		}
		else
		{
			[self errorMessage:S_MOVEERRORTITLE text:[NSString stringWithFormat:S_MOVEERRORCONTENT,
			[filename lastPathComponent],[destination stringByDeletingLastPathComponent]]];
		}
	}
	else
	{
		if([[NSFileManager defaultManager] copyPath:filename toPath:destination handler:nil])
		{
			// "copied" message in status bar
			[[[[NSSound alloc] initWithContentsOfFile:@"/System/Library/Components/CoreAudio.component/Contents/Resources/SystemSounds/system/Volume Mount.aif" byReference:NO] autorelease] play];
		}
		else
		{
			[self errorMessage:S_COPYERRORTITLE	text:[NSString stringWithFormat:S_COPYERRORCONTENT,
			[filename lastPathComponent],[destination stringByDeletingLastPathComponent]]];
		}
	}
}

-(IBAction)jpegAutoRotate:(id)sender { [self jpegTransform:@"-a"]; }

-(IBAction)jpegRotateCW:(id)sender { [self jpegTransform:@"-9"]; }

-(IBAction)jpegRotateCCW:(id)sender { [self jpegTransform:@"-2"]; }

-(IBAction)jpegRotate180:(id)sender { [self jpegTransform:@"-1"]; }

-(IBAction)jpegFlipHorizontal:(id)sender { [self jpegTransform:@"-F"]; }

-(IBAction)jpegFlipVertical:(id)sender { [self jpegTransform:@"-f"]; }

-(void)jpegTransform:(NSString *)option
{
	if(currindex<0) return;
	if(!currimage) return;
	if(![[currimage format] isEqual:@"JPEG"]) return;

	NSString *filename=[self currentFilename];
	NSTask *task=[NSTask launchedTaskWithLaunchPath:[[[NSBundle mainBundle] resourcePath] stringByAppendingPathComponent:@"exiftran"]
	arguments:[NSArray arrayWithObjects:@"-ip",option,filename,nil]];

	[task waitUntilExit];

	if([task terminationStatus])
	{
		[self errorMessage:S_JPEGERRORTITLE text:[NSString stringWithFormat:S_JPEGERRORCONTENT,filename]];
	}
	[self reloadImage];
}

-(IBAction)launchAppFromMenu:(id)sender
{
	if(currindex<0) return;

	NSString *filename=[self currentFilename];
	NSString *app=[sender representedObject];

	[[NSWorkspace sharedWorkspace] openFile:filename withApplication:app];
}

-(IBAction)launchDefaultEditor:(id)sender
{
	if(currindex<0) return;

	NSString *filename=[self currentFilename];
	NSString *app=[(XeeDelegate *)[[NSApplication sharedApplication] delegate] defaultEditor];

	[[NSWorkspace sharedWorkspace] openFile:filename withApplication:app];
}



-(void)errorMessage:(NSString *)title text:(NSString *)text
{
	NSAlert *alert=[[[NSAlert alloc] init] autorelease];

	[alert setMessageText:title];
	[alert setInformativeText:text];
	[alert addButtonWithTitle:S_OK];

	[alert beginSheetModalForWindow:window modalDelegate:nil didEndSelector:nil contextInfo:nil];
}



+(void)initialize
{
	if(!directorycontrollers) directorycontrollers=[[NSMutableArray alloc] initWithCapacity:16];
}

+(NSArray *)controllers { return directorycontrollers; }


@end



@implementation XeeDirectoryWindow

static float minstep=1e10;

-(void)keyDown:(NSEvent *)event
{
	unichar c=[[event characters] characterAtIndex:0];

	if(c==27) [[NSApplication sharedApplication] sendAction:@selector(performClose:) to:nil from:self];
	else if(c=='=') [[NSApplication sharedApplication] sendAction:@selector(zoomIn:) to:nil from:self];
	else if(c==',') [[NSApplication sharedApplication] sendAction:@selector(skipPrev:) to:nil from:self];
	else if(c=='.') [[NSApplication sharedApplication] sendAction:@selector(skipNext:) to:nil from:self];
	else if(c==' ') [[NSApplication sharedApplication] sendAction:@selector(skipNext:) to:nil from:self];
	else if(c==127) [[NSApplication sharedApplication] sendAction:@selector(askAndDelete:) to:nil from:self];
	else if(c==63272) [[NSApplication sharedApplication] sendAction:@selector(askAndDelete:) to:nil from:self];
	else [super keyDown:event];

	//printf("key down: %d\n",c);
}

-(void)scrollWheel:(NSEvent *)event
{
//	[super scrollWheel:event];

	float step=[event deltaY];
	float absstep=fabsf(step);
	if(absstep<minstep) minstep=absstep;
	if(minstep<1) minstep=1;

	int n=-(int)((absstep+0.5)/minstep);
	if(signbit(step)) n=-n;

	[controller skip:n];
}

@end
